/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/* @(#)16       1.104.1.6  src/avs/fs/mmfs/ts/kernext/gpl-linux/cxiSystem.c, mmfs, avs_rcs2b, rcs2bs004a 6/1/04 08:55:17 */
/*
 * Linux implementation of basic common services
 *
 * Contents:
 *     cxiGetThreadId
 *     getpid
 *     cxiIsSuperUser
 *     DoPanic
 *     logAssertFailed
 *   Kernel memory allocation services:
 *     cxiMallocPinned
 *     cxiFreePinned
 *
 */

#include <Shark-gpl.h>

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <linux/time.h>
#include <linux/file.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#undef memcmp

#define DEFINE_TRACE_GBL_VARS
#include <verdep.h>
#include <linux2gpfs.h>
#include <cxiSystem.h>
#include <cxiAtomic.h>
#include <cxi2gpfs.h>
#include <cxiIOBuffer.h>
#include <cxiSharedSeg.h>
#include <cxiCred.h>

#include <Trace.h>
#include <lxtrace.h>
#include <cxiMode.h>

#if LINUX_KERNEL_VERSION >= 2040900
/* This is in the Redhat kernel series */
extern int posix_locks_deadlock(struct file_lock *, struct file_lock *);
#endif

#ifdef INSTRUMENT_LOCKS
struct BlockingMutexStats BlockingMutexStatsTable[MAX_GPFS_LOCK_NAMES];
#endif  /* INSTRUMENT_LOCKS */

/* We record the daemon's process group since it can uniquely identify
 * a thread as being part of the GPFS daemon.  pid is unique per thread
 * on linux due to their clone implementation.
 */
static pid_t DaemonPGrp = -1;

/* Get the kernel thread ID. */
cxiThreadId cxiGetThreadId()
{
  return current->pid;
}

/* Get the kernel process ID. */
pid_t getpid()
{
  return current->pid;
}

/* bufP is caller's ext_cred_t buffer
 * uCredPP is the ucred struct (NULL on Linux)
 * eCredPP is the ext_cred_t struct * (if successful)
 * 
 * cxiPutCred should be called to release when operation has been completed.
 */
int cxiGetCred(void *bufP, void **uCredPP, void **eCredPP)
{
  ext_cred_t *eCredP = (ext_cred_t *)bufP;

  *uCredPP = NULL;
  *eCredPP = NULL;

  if (!bufP)
    return EINVAL;

  setCred(eCredP);
  *eCredPP = (void *)eCredP;

xerror:
  return 0;
}

/* Release of cxiGetCred() structures  (nothing to do on Linux) */
int cxiPutCred(void *userCredP, void *extCredP)
{
  if (userCredP || !extCredP)
    return EINVAL;

  return 0;
}

Boolean
cxiIsKswapdThread()
{
  return (strcmp(current->comm, "kswapd") == 0);
}

/* Convert a kernel stack address to the thread ID of the thread that 
 * uses that stack 
 */
int 
cxiStackAddrToThreadId(char* stackP, cxiThreadId* tidP)
{
  struct task_struct * tP;
#if LINUX_KERNEL_VERSION >= 2060000
  /* the kernel stack is base off the thread_info struct in the 2.6 kernel
   * will get the task pointer out of thread_info struct.
   */
  struct thread_info * iP;
  iP = (struct thread_info *) ((UIntPtr)stackP & ~((UIntPtr)(THREAD_SIZE-1)));
  tP=iP->task;
#else
  /* the kernel stack is base off the task_struct struct in the 2.4 kernel */
  tP = (struct task_struct *) ((UIntPtr)stackP & ~((UIntPtr)(THREAD_SIZE-1)));
#endif
  *tidP = tP->pid;
  return 0;
}

/* Convert a kernel thread pointer to the corresponding thread ID */
int 
cxiThreadPtrToThreadId(char* threadP, cxiThreadId* tidP)
{
  struct task_struct * tP;

  tP = (struct task_struct *) threadP;
  *tidP = tP->pid;

  return 0;
}


/* Return true if caller has has maximum authorization (is root) */
Boolean cxiIsSuperUser()
{
  return (current->euid == 0);
}


/* Get the process max filesize limit (ulimit -f) */
Int64 cxiGetMaxFileSize()
{
  if ((signed long)current->rlim[RLIMIT_FSIZE].rlim_cur == -1L)
    return MAX_INT64;
  else
    return (current->rlim[RLIMIT_FSIZE].rlim_cur);
}

/* Routine to send a signal to the current thread/process */
void cxiSendSigThread(int sig)
{
  send_sig(sig, current, 0);
}


#ifdef MALLOC_DEBUG
/* This tracks mallocs and frees on a limited basis.
 * Implemented originally to determine if we were leaking
 * any memory after an unload.  This is not really thread
 * safe for multiple processors unless they're automatically
 * cache coherent without memory barriers (i386).   Its useful
 * for detecting memory leaks on a single processor system.
 */
#define MALLOC_RECORDS 5000 /* max mallocs to track */
struct mallocStat
{
  void *beginP;
  unsigned short size;
  unsigned short type;
};
static struct mallocStat *mstatP = NULL;
static unsigned int mstatEnd = 0;

void 
MallocDebugStart()
{
  int i;

  if (mstatP == NULL)
    mstatP = vmalloc(MALLOC_RECORDS * sizeof(struct mallocStat));
  
  if (mstatP == NULL)
    return;

  for (i = 0; i < MALLOC_RECORDS; i++)
  {
    mstatP[i].beginP = NULL;  
    mstatP[i].size = 0;
    mstatP[i].type = 0;
  }
  printk("MallocDebugStart 0x%X\n", mstatP);
}

void
MallocDebugEnd()
{
  int i;

  if (mstatP != NULL)
  {
    for (i = 0; i < mstatEnd && i < MALLOC_RECORDS; i++)
    {
      if (mstatP[i].beginP != NULL)
        printk("MallocDebug: beginP 0x%X size %d type %d index %d\n",
               mstatP[i].beginP, mstatP[i].size, mstatP[i].type, i);
    }
  }

  vfree(mstatP);
  mstatEnd = 0;
  mstatP = NULL;
}

static void
MallocDebugNew(void *ptr, unsigned short size, unsigned short type)
{
  int oldval;

  oldval = ATOMIC_ADD(&mstatEnd, 1);
  if (oldval > MALLOC_RECORDS || mstatP == NULL)
   return;

  if (mstatP[oldval].beginP != NULL)
    printk("??? oldval %d beginP 0x%X mstatEnd %d\n", 
           oldval, mstatP[oldval].beginP, mstatEnd);

  mstatP[oldval].beginP = ptr;
  mstatP[oldval].size = size;
  mstatP[oldval].type = type;
}

static void 
MallocDebugDelete(void *ptr)
{
  int i;
  int found = 0;
 
  if (mstatP == NULL)
    return;

  for (i = 0; i < mstatEnd && i < MALLOC_RECORDS; i++)
    if (mstatP[i].beginP == ptr)
    {
      mstatP[i].beginP = NULL;
      mstatP[i].size = 0;
      mstatP[i].type = 0;
      found = 1;
      break;
    }

  if (!found)
    printk("MallocDebug: 0x%X not found! mstatEnd %d\n", ptr, mstatEnd);
}
#endif /* MALLOC_DEBUG */

/* Allocate pinned kernel memory */
void* cxiMallocPinned(int nBytes)
{
  void *ptr;

  /* kmalloc only supports requests for up to 131027 bytes.  Anything
     larger than this results in a BUG() call. */
  if (nBytes > 131072)
    return NULL;

  ptr = kmalloc(nBytes, GFP_KERNEL);

#ifdef MALLOC_DEBUG
  MallocDebugNew(ptr, nBytes, 1);
#endif

  return ptr;
}

/* Free pinned kernel memory that was allocated with cxiMallocPinned */
/* Must not block on lack of memory resourses */
void cxiFreePinned(void* p)
{
#ifdef MALLOC_DEBUG
  MallocDebugDelete(p);
#endif

  kfree(p);
}

/* Get the kernel thread ID. */
void* cxiGetFcntlOwner(eflock_t *flP)
{
  return flP? flP->l_owner: current->files;
}

/* Perform local advisory locking. */
int cxiFcntlLock(void *advObjP,
                 int cmd,
                 void *lockStructP,
                 cxiFlock_t *flockP,
                 int (*retryCB)(),
                 cxiOff64_t size,
                 cxiOff64_t offset,
                 ulong *retry_idP)
{
  int rc = 0;
  // struct file *fP;
  struct file_lock fl, *flP, *gflP, *cflP;
  Boolean keepLockElement = false;

  /* cast platform independent arguments as appropriate for linux */
  void (*RetryFcn)(struct file_lock*) = (void (*)(struct file_lock*))retryCB;
  // fP = (struct file *)advObjP;
  struct file localFile;
  struct dentry localDEntry;
  flP = (struct file_lock *) lockStructP;

  localFile.f_dentry = &localDEntry;
  localDEntry.d_inode = (struct inode *)advObjP;

  /* Callers have the option of passing a platform dependent lock structure
     (struct file_lock *lockSructP) or the generic (cxiFlock_t *flockP). */
  if (flockP)
  {
    flP = &fl; /* Use a local file_lock structure */

    /* If there is a potential for blocking, must malloc the locking structure
       so it can persist until the lock becomes available (in Retry()). */

    if (cmd == F_SETLKW)
    {
      flP = (struct file_lock*)cxiMallocUnpinned(sizeof(struct file_lock));
      if (flP == NULL)
      {
	rc = ENOMEM;
	goto exit;
      }
    }

    cxiMemset(flP, 0, sizeof(*flP));
    locks_init_lock(flP); /* Initialize list_head structs */

    flP->fl_flags = FL_POSIX;
    flP->fl_file  = NULL; /* don't pass stack file struct */

    /* fl_wait needs to be initialized because when unlock happens, the
       linux routine locks_wake_up_blocks invokes our retry routine via
       fl_notify and then calls wake_up(fl_wait) on the assumption that
       the waiter is local. */

    cxiWaitEventInit((cxiWaitEvent_t *)&flP->fl_wait);

    cxiFlockToVFS(flockP, flP);
  }

  /* daemon didn't know the owner and required kernel code to fill it in. */
  if (!flP->fl_owner)
    flP->fl_owner = (fl_owner_t)cxiGetFcntlOwner(NULL);

#if 0
  /* Validate the file pointer.  Kernel locking routines are going to
     use these without verifying them.  If any of them are NULL, find
     out now before they generate a segment violation. */
  if ((!fP) || (!fP->f_dentry) || (!fP->f_dentry->d_inode))
  {
    if (cmd == F_GETLK)
      flP->fl_type = F_UNLCK;
    else
      rc = EINVAL;
    goto exit;
  }
#endif

  /* Note that this all depends on us having serialized such locking for
     this file during from before the posix_test_lock() until after the
     posix_block_lock().  The revoke lock that we hold here provides us
     the necessary serilization. */

  TRACE7(TRACE_VNODE, 3, TRCID_FCNTLLOCK_ENTER,
         "cxiFcntlLock posix_lock_file: pid %d owner 0x%X inodeP 0x%X "
         "range 0x%lX-%lX cmd %s type %s\n",
         flP->fl_pid, flP->fl_owner, advObjP, flP->fl_start, flP->fl_end,
         (cmd == F_GETLK) ? "GETLK" : (cmd == F_SETLK) ? "SETLK" : "SETLKW",
         (flP->fl_type == F_RDLCK) ? "RDLCK" :
         (flP->fl_type == F_WRLCK) ? "WRLCK" : "UNLCK");

  if (cmd == F_GETLK)
  {
    /* Check for conflicts.  If found, return the information.
       If there are NO conflicts, return F_UNLCK in fl_type. */
    if (NULL != (gflP = posix_test_lock(&localFile, flP)))
    {
      flP->fl_start = gflP->fl_start;
      flP->fl_end = gflP->fl_end;
      flP->fl_type = gflP->fl_type;
      flP->fl_pid = gflP->fl_pid;
      flP->fl_owner = gflP->fl_owner;
    }
    else
      flP->fl_type = F_UNLCK;

    TRACE6(TRACE_VNODE, 3, TRCID_FCNTLLOCK_GETLK,
           "cxiFcntlLock getlk: pid %d owner 0x%X inodeP 0x%X "
           "range 0x%lX-%lX type %s\n",
           flP->fl_pid, flP->fl_owner, advObjP, flP->fl_start, flP->fl_end,
           (flP->fl_type == F_RDLCK) ? "RDLCK" :
           (flP->fl_type == F_WRLCK) ? "WRLCK" : "UNLCK");
  }
  else
  { /* Begin: do the locking, but handle the blocking via our retry routine. */
    /* Test the lock.   What this really does for us is return the blocker
       if one exists.  This is needed to queue up the request if a conflicting
       lock is already held. */

    if ((flP->fl_type == F_UNLCK) || !(cflP = posix_test_lock(&localFile, flP)))
    {
      /* No conflicting lock:  get the lock for the caller. */
      rc = POSIX_LOCK_FILE(&localFile, flP);
    }
    else
    { /* Conflicting lock:  ..... */
      rc = EAGAIN;

      if (cmd == F_SETLKW)
      {
        if (posix_locks_deadlock(flP, cflP))
        {
          rc = EDEADLK;
        }
        else
        {
          /* Queue the blocker structures */
          keepLockElement = true;
          if (retry_idP)
            *retry_idP = (ulong)flP; // returned to caller and saved in sleepElement
          flP->fl_notify = RetryFcn;
          posix_block_lock(cflP, flP);
        }
      }
    }

    TRACE2(TRACE_VNODE, 3, TRCID_FCNTLLOCK_EXIT,
           "cxiFcntlLock posix_lock_file: rc 0x%X retry_id 0x%lX\n", rc, cflP);
  } /* End: do the locking, but handle the blocking via our retry routine. */

exit:

  if (flockP)
  {
    /* Caller wanted results in flockP */
    cxiVFSToFlock((void *)flP, flockP);

    /* If we allocated the locking structure and then didn't need to use
       it (the lock request didn't block), free it. */

    if ((flP!=&fl) && (!keepLockElement))
      cxiFreeUnpinned(flP);
  }

  return rc;
}

void cxiFcntlUnblock(void *retry_idP)
{
  struct file_lock *flP = (struct file_lock *)retry_idP;

  /* Include some sanity checks on the retry id (file_lock)
     before passing it into the routine that does the work.
     It should be properly linked (via its list_head structures)
     in a file_lock_list that has blocked waiters.  Also,
     we would only be backing this out by the process that
     has originally blocked, so verify the pid. */

  if (!list_empty(&flP->fl_block) && !list_empty(&flP->fl_link)  &&
       flP->fl_next && flP->fl_pid == getpid())
  {
    POSIX_UNBLOCK_LOCK(flP);
  }
}

int
cxiFcntlReset(void *vfsP, cxiPid_t mmfsd_pid)
{
  int rc = 0;
  struct super_block *sbP = (struct super_block *)vfsP;
  struct list_head *fllP;
  struct file_lock *fl;
  struct dentry *dentryP;

  lock_kernel();

restart:
  fllP = file_lock_list.next;

  while(fllP != &file_lock_list)
  {
    fl = list_entry(fllP, struct file_lock, fl_link);
    fllP = fllP->next;

    /* If there are mmfs lock structures, release them. */

    if (fl &&
        fl->fl_file &&
        fl->fl_file->f_dentry &&
        fl->fl_file->f_dentry->d_inode)
    {
      dentryP = fl->fl_file->f_dentry;

      /* If this lock belongs to the specified vfs, release advisory locks. */
      if (dentryP->d_sb == sbP)
      {
        /* remove all our locks */
        rc = gpfs_ops.gpfsFcntlReset((void *)dentryP->d_inode, mmfsd_pid);
        if (rc == ENOSYS)
          goto xerror;

        /* After freeing unknown numbers of locks in gpfsFcntlReset (all
           locks for the inode), restart from the top of the lock list */
        goto restart;
      }
    }
  }

xerror:
  unlock_kernel();
  return rc;
}

void *
cxiGetPrivVfsP(void *vfsP)
{
  struct super_block *sbP = (struct super_block *)vfsP;
  return (SBLOCK_PRIVATE(sbP));
}


#ifdef NFS_DEBUG
/* These flags are defined in the kernel and control various cprintk
   calls.  This provides us a way to easily turn these on/off for
   debugging our NFS support. */
extern unsigned int nlm_debug;
extern unsigned int nfsd_debug;
extern unsigned int nfs_debug;
extern unsigned int rpc_debug;
#endif

int cxiTrace(cxiTrace_t trace)
{
#ifdef NFS_DEBUG
  int rc = 0;

  switch (trace)
  {
    case cxiTraceNFS:
      nlm_debug = nfsd_debug = nfs_debug = rpc_debug = ~0;
      break;
    case cxiTraceNFSoff:
      nlm_debug = nfsd_debug = nfs_debug = rpc_debug =  0;
      break;
    default:
      rc = EINVAL;
      break;
  }
  return rc;
#else
  return ENOSYS;
#endif
}

void cxiFlockToVFS(eflock_t* lckdatP, void* vP)
{
  struct file_lock* flP = (struct file_lock *)vP;

  if ((flP) && (lckdatP))
  {
    flP->fl_pid   = lckdatP->l_pid;
    flP->fl_owner = lckdatP->l_owner;
    flP->fl_type  = lckdatP->l_type;
    flP->fl_start = lckdatP->l_start;
    flP->fl_flags = FL_POSIX;
    if (lckdatP->l_caller == L_CALLER_LOCKD)
      flP->fl_flags |= FL_LOCKD;
    if (lckdatP->l_len == 0)
      flP->fl_end = FL_OFFSET_MAX;
    else
      flP->fl_end = lckdatP->l_len + lckdatP->l_start - 1;
  }
  return;
}

void cxiVFSToFlock(void *vP, eflock_t *lckdatP)
{
  struct file_lock* flP = (struct file_lock *)vP;

  if ((flP) && (lckdatP))
  {
    lckdatP->l_pid    = flP->fl_pid;
    lckdatP->l_owner  = flP->fl_owner;
    lckdatP->l_type   = flP->fl_type;
    lckdatP->l_start  = flP->fl_start;
    if (flP->fl_flags & FL_LOCKD)
      lckdatP->l_caller = L_CALLER_LOCKD;
    else
      lckdatP->l_caller = L_CALLER_NULL;
    if (flP->fl_end == FL_OFFSET_MAX)
      lckdatP->l_len = 0;
    else
      lckdatP->l_len    = flP->fl_end - flP->fl_start + 1;
  }
  return;
}


/* Sleep for the indicated number of milliseconds */
void cxiSleep(int ms)
{
  TRACE1(TRACE_VNODE, 9, TRCID_SLEEP,
         "cxiSleep: begin delay %d\n", ms);
  current->state = TASK_INTERRUPTIBLE;
  /* For large HZ rearrange jiffies calculation and
     use presumably larger word size to minimize overflow risk */
  if (unlikely(HZ > 1000))
    schedule_timeout(((long)ms)*HZ/1000);
  else
    schedule_timeout(ms/(1000/HZ));
  TRACE2(TRACE_VNODE, 9, TRCID_SLEEP_END,
         "cxiSleep: end delay %d HZ %d\n", ms, HZ);
}


void cxiOpenNFS(void *iP)
{
  struct inode *inodeP = (struct inode *)iP;
  int refcount;

  /* A reference is placed on the cxiNode here when the first NFS reference
     is added */
  refcount = cxiRefOSNode(NULL, ((cxiNode_t *)(cxiGetCnP(inodeP))), iP, 1);

  TRACE7(TRACE_VNODE, 3, TRCID_OPENNFS,
        "openNFS iP 0x%lX ino %d (0x%X) mode 0x%X nlink %d gen_ip 0x%lX "
        "refcount %d\n",
        inodeP, (inodeP) ? inodeP->i_ino : -1,
        (inodeP) ? inodeP->i_ino : -1,
        (inodeP) ? inodeP->i_mode : -1,
        (inodeP) ? inodeP->i_nlink : -1,
        (inodeP) ? inodeP->u.generic_ip : NULL,
        refcount);

  DBGASSERT(refcount != 0);
}


int cxiCloseNFS(void *vP, void *viP)
{
  int rc;
  struct inode *iP = (struct inode *)vP;

  /* If viP is NULL, the file was never actually opened.
     If viP is not NULL, close it. */
  if (viP == NULL)
    rc = 0;
  else
    rc = gpfs_ops.gpfsClose(VP_TO_PVP(iP), VP_TO_CNP(iP), FREAD|FWRITE,
                            (struct MMFSVInfo *)viP, true);

  cxiPutOSNode((void *)iP);

  return rc;
}

void * cxiGetNfsP(void *vP)
{
  if (vP && VP_TO_CNP((struct inode *)vP))
    return VP_TO_NFSP((struct inode *)vP);
  else
    return NULL;
}

void cxiSetNfsP(void *vP, void *newP)
{
  if (VP_TO_CNP((struct inode *)vP))
    VP_TO_NFSP((struct inode *)vP) = newP;
}

void * cxiGetCnP(void *vP)
{ return (void *)VP_TO_CNP((struct inode *)vP); }

void * cxiGetPvP(void *vP)
{ return (void *)VP_TO_PVP((struct inode *)vP); }

void * cxiGNPtoVP(void *vP)
{ return (void *)GNP_TO_VP((struct cxiNode_t *)vP); }

/* Main routine of kproc */
static int kprocMain(void *argP)
{
  cxiKProcData_t *kpdP = (cxiKProcData_t *)argP;

  /* Change our process name */
  current->comm[sizeof(current->comm) - 1] = '\0';
  strncpy(current->comm, kpdP->nameP, sizeof(current->comm) - 1);

  /* Change parent of a kernel process so that when it exits, it won't 
   * send a SIGCHLD signal to the process that created it, and it won't 
   * be left as a zombie. 
   */
  DAEMONIZE(kpdP->nameP);

  /* Call the function specified by startKProc */
  kpdP->func(kpdP);
  return 0;
}

/* Create a new kernel process */
cxiPid_t 
cxiStartKProc(struct cxiKProcData_t *kpdP)
{
  cxiPid_t pid = kernel_thread(kprocMain, kpdP, kpdP->kprocFlags);
  kpdP->pid = pid > 0 ? pid : KPROC_FAILED_PID;

  TRACE2(TRACE_VNODE, 1, TRCID_CXISTART_KPROC_LINUX,
         "cxiStartKProc %s pid %d \n", kpdP->nameP, kpdP->pid);
  return kpdP->pid;
}

void 
cxiStopKProc(struct cxiKProcData_t *kpdP)
{
  cxiPid_t pid;

  cxiBlockingMutexAcquire(&kpdP->lock);
  
  TRACE2(TRACE_VNODE, 1, TRCID_CXISTOP_KPROC_LINUX,
         "cxiStopKProc: %s pid %d \n", kpdP->nameP, kpdP->pid);

  if (!KPROC_RUNNING(kpdP))
  {
    cxiBlockingMutexRelease(&kpdP->lock);
    return;
  }

  pid = kpdP->pid;              // Cache pid before signal/wait
  kpdP->terminate = true;
  cxiWaitEventSignal(&kpdP->kprocEvent);

  while (kpdP->pid != KPROC_UNASSIGNED_PID)
    cxiWaitEventWait(&kpdP->startStopEvent, &kpdP->lock, 0);

  cxiBlockingMutexRelease(&kpdP->lock);
}

/*-------------------------------------------------------------------
 * logAssertFailed  - Subroutine consolidating logGenIF() and
 *                    DoPanic() calls.
 *------------------------------------------------------------------*/

static char PanicMsgBuf[2048];

void cxiPanic(const char* panicStrP)
{
  printk( GPFS_NOTICE  "kp %d: cxiPanic: %s\n", cxiGetThreadId(), panicStrP);
  TRACE1(TRACE_ERRLOG, 0, TRCID_PANIC, "cxiPanic: %s\n", panicStrP);
#ifndef DISABLE_KERNEL_PANIC
  BUG();
#endif
}

static void
DoPanic(char* condP, char* filenameP, int lineNum, Int32 retCode,
        Int32 reasonCode, char *dataStr)
{
  const char *p;
  int bytesLeft;

  p = cxiStrrchr(filenameP, '/');
  if (p == NULL)
    p = filenameP;
  else
    p += 1;

  sprintf(PanicMsgBuf, "%s:%d:%d:%d:", p, lineNum, retCode, reasonCode);
  bytesLeft = sizeof(PanicMsgBuf) - strlen(PanicMsgBuf);
  if (dataStr)
  {
    strncat(PanicMsgBuf, dataStr, bytesLeft-1);
    bytesLeft = sizeof(PanicMsgBuf) - strlen(PanicMsgBuf);
  }
  strncat(PanicMsgBuf, ":", bytesLeft-1);
  bytesLeft = sizeof(PanicMsgBuf) - strlen(PanicMsgBuf);
  if (condP)
    strncat(PanicMsgBuf, condP, bytesLeft-1);
  cxiPanic(PanicMsgBuf);
}

#ifdef MODULE
void
logAssertFailed(UInt32 flags,   /* LOG_FATAL_ERROR or LOG_NONFATAL_ERROR */
                char  *srcFileName,   /* __FILE__ */
                UInt32 srcLineNumber, /* __LINE__ */
                Int32  retCode,       /* return code value */
                Int32  reasonCode,    /* normally errno */
                UInt32 logRecTag,     /* tag if have associated error log rec */
                char  *dataStr,       /* assert data string */
                char  *failingExpr)   /* expression that evaluated to false */
{
  int i;

  printk("GPFS logAssertFailed: %s file %s line %d\n",
         failingExpr, srcFileName, srcLineNumber);
  TRACE3(TRACE_ERRLOG, 0, TRCID_MODULE_LOGASSERT_1,
         "logAssertFailed: %s retCode %d reasonCode %d\n",
         failingExpr, retCode, reasonCode);
  TRACE2(TRACE_ERRLOG, 0, TRCID_MODULE_LOGASSERT_2,
         "logAssertFailed: file %s line %d\n", srcFileName, srcLineNumber);
#ifndef GPFS_PRINTF
  /* fsync buffered lxtrace records */
  trc_fsync();

#ifdef STOP_TRACE_ON_FAILURE
  /* Turn off tracing right after the failure occurs.  This may only turn
     off tracing in the kernel. */
  for (i=0 ; i<MAX_TRACE_CLASSES ; i++)
    TraceFlagsP[i] = 0;
#endif

  /* Wait 10 seconds to allow the lxtrace daemon to complete the sync. */
  cxiSleep(10000);
#endif
  gpfs_ops.gpfsDaemonToDie(srcFileName, srcLineNumber, retCode, reasonCode,
                           dataStr, failingExpr);

  DoPanic(failingExpr, srcFileName, srcLineNumber, retCode, reasonCode,
          dataStr);
}
#else /* !MODULE */
void
logAssertFailed(UInt32 flags,
                char  *srcFileName,
                UInt32 srcLineNumber,
                Int32  retCode,
                Int32  reasonCode,
                UInt32 logRecTag,
                char  *dataStr,
                char  *failingExpr);
#endif /* MODULE */


typedef struct cxiWaitElement_t
{
  cxiWaitList_t waitList;  /* previous and next element in chain */

  /* Linux would normally organize a wait_queue_head_t with any number
   * of wait_queue_t elements.  However since we're implementing "wakeup
   * with return code" we have to ensure the OS wakes up the exact sleeper
   * we want.  Thus we have only a one to one relationship to ensure the
   * OS can only pick our favorite.
   */
  wait_queue_head_t qhead;
  wait_queue_t qwaiter;
  int wakeupRC;            /* wakeup return code */

} cxiWaitElement_t;


#define CXI_WAIT_LIST_ADD(headP, elementP) \
   (headP)->prevP->nextP = (elementP); \
   (elementP)->prevP = (headP)->prevP; \
   (headP)->prevP = (elementP);        \
   (elementP)->nextP = (headP);

#define CXI_WAIT_LIST_REMOVE(elementP) \
   (elementP)->prevP->nextP = (elementP)->nextP; \
   (elementP)->nextP->prevP = (elementP)->prevP;


/* Initialize abstract wait event with OS specific
 * initialization function
 */
void
cxiWaitEventInit(cxiWaitEvent_t *weP)
{
  spinlock_t *lockP = (spinlock_t *)&weP->lword;

  spin_lock_init(lockP);
  weP->waitList.nextP = weP->waitList.prevP = &weP->waitList;
}

Boolean
cxiWaitEventHasWaiters(cxiWaitEvent_t *weP)
{
  unsigned long flags;
  spinlock_t *lockP = (spinlock_t *)(weP->lword);
  Boolean rc;

  SPIN_LOCK_IRQ(lockP, flags);
  rc = (weP->waitList.nextP != &weP->waitList);
  SPIN_UNLOCK_IRQ(lockP, flags);
  return rc;
}

/* Do not add trace records.  Some callers depend on not being
 * interrupted by the trace daemon.
 */
enum WakeType { wBroadcast, wSignal, wWakeOne };
static inline void
doWakeup(cxiWaitEvent_t *wEventP, enum WakeType wtype, int wakeupRC)
{
  unsigned long flags;
  spinlock_t *lockP = (spinlock_t *)(wEventP->lword);
  cxiWaitList_t *headP;
  cxiWaitList_t *tmpP;
  cxiWaitElement_t *wP;

  SPIN_LOCK_IRQ(lockP, flags);

  /* We wake up from the front back (FIFO semantics).
   * There's only one wait element per wake_queue_head_t so
   * record the return code and wake up the one element.
   */
  headP = &wEventP->waitList;

  for (tmpP = headP->nextP; tmpP != headP; tmpP = tmpP->nextP)
  {
    wP = list_entry(tmpP, cxiWaitElement_t, waitList);
    wP->wakeupRC = wakeupRC;

    wake_up(&wP->qhead);
    if (wtype != wBroadcast)
    {
      /* The difference between wSignal and wWakeOne is that the latter
         guarantees that multiple wake up calls will each pick a different
         thread if more than one is waiting.  With wSignal, if a thread is
         awakened but hasn't had a chance to run, then subsequent wake up
         calls might all wake the same thread.

         On AIX, the calling routine (e_wakeup_one) removes the waiter from
         the queue, unlike Linux where removal is done by the waiting
         thread when it wakes up.  Nothing special has to be done on AIX to
         get the nWakeOne style of wakeup. 

         Note:  This is an inline routine and the wType argument is a
         compile-time constant, so the "if" tests in this routine are done
         by the compiler and do not generate any code. */

      if (wtype == wWakeOne)
      {
        /* Move this entry to tail of list so that the next wakeup call will
           pick somebody else. */
        CXI_WAIT_LIST_REMOVE(tmpP);
        CXI_WAIT_LIST_ADD(headP, tmpP);
      }
      break;
    }
  }
  SPIN_UNLOCK_IRQ(lockP, flags);
}

int
cxiCopyIn(char *from, char *to, unsigned long size)
{
  /* The daemon needs to bypass access checks since copy to 
   * shared segment would inadvertantly fail.
   */
  if (PROCESS_GROUP(current) == DaemonPGrp)
    __copy_from_user(to, from, size);
  else 
    if (copy_from_user(to, from, size))
      return EFAULT;
  
  return 0;
}

int
cxiCopyOut(char *from, char *to, unsigned long size)
{
  /* The daemon needs to bypass access checks since copy to 
   * shared segment would inadvertantly fail.
   */
  if (PROCESS_GROUP(current) == DaemonPGrp)
    __copy_to_user(to, from, size);
  else
    if (copy_to_user(to, from, size))
      return EFAULT;
  
  return 0;
}

int
cxiCopyInstr(char *from, char *to, unsigned long size, unsigned long *len)
{
  long retval;
  retval = strncpy_from_user(to, from, size);
  if ((retval > 0) && (retval <= size))
  {
    *len = retval;
    return 0;
  }
  *len = 0;
  if (retval < 0)
    retval = EFAULT;
  else
    retval = ENAMETOOLONG;
  return (int)retval;
}

#ifdef GPFS_ARCH_X86_64
/* Check if 64-bit user process */
int
cxiIS64U(char *addr)
{
  return !(current->thread.flags & THREAD_IA32);
}
#endif

int 
socket_aio_dequeue()
{
  return -1;
}

/* Transfer data from buffer(s) in user space to or from a buffer in the
   kernel. */
int
cxiUiomove(register char* kBufP,          /* address of kernel buffer */
           register unsigned long nBytes, /* #bytes to transfer */
           Boolean toKernel,              /* direction of xfer(read/write)*/
           register struct cxiUio_t* uioP) /* user area description */
{
  register struct cxiIovec_t * iovP;
  unsigned long cnt;
  int rc;
#ifdef TRACE_IO_DATA
  char* origKBufP = kBufP;
  int trcdata[4];
#endif

  TRACE4(TRACE_FOPS, 2, TRCID_CXISYSTEM_037,
         "cxiUiomove enter: kBufP 0x%lX uioP 0x%lX nBytes %d toKernel %d\n",
         kBufP, uioP, nBytes, toKernel);
  if (uioP->uio_resid <= 0)
    return ENOMEM;
  rc = 0;
  if (uioP->uio_iovcnt == 1)
  {
    /*
     * Fastpath for most common case of iovcnt == 1.  Saves a
     * few instructions.
     */
    iovP = uioP->uio_iov;
    cnt = iovP->iov_len;
    if (cnt <= 0)
    {
      uioP->uio_iovcnt--;
      uioP->uio_iov++;
      uioP->uio_iovdcnt++;
      return 0;
    }
    if (cnt > nBytes)
      cnt = nBytes;

    if (toKernel)
    {
      /* The daemon needs to bypass access checks since copy to 
       * shared segment would inadvertantly fail.  Copies to 
       * kernel address space also perform no validity check.
       */
      if (PROCESS_GROUP(current) == DaemonPGrp || uioP->uio_segflg == UIO_SYSSPACE)
        __copy_from_user(kBufP, (char *)iovP->iov_base, cnt);
      else
        if (copy_from_user(kBufP, (char *)iovP->iov_base, cnt))
          return EFAULT;
    }
    else
    {
      /* The daemon needs to bypass access checks since copy to 
       * shared segment would inadvertantly fail.  Copies to 
       * kernel address space also perform no validity check.
       */
      if (PROCESS_GROUP(current) == DaemonPGrp || uioP->uio_segflg == UIO_SYSSPACE)
        __copy_to_user((char *)iovP->iov_base, kBufP, cnt);
      else
        if (copy_to_user((char *)iovP->iov_base, kBufP, cnt))
          return EFAULT;
    }

    iovP->iov_base = (char *)iovP->iov_base + cnt;
    iovP->iov_len -= cnt;
    uioP->uio_resid -= cnt;
    uioP->uio_offset += cnt;
#ifdef TRACE_IO_DATA
    if (cnt >= sizeof(trcdata))
      memcpy(trcdata, origKBufP, sizeof(trcdata));
    else
    {
      memset(trcdata, 0xAA, sizeof(trcdata));
      memcpy(trcdata, origKBufP, cnt);
    }
    TRACE5(TRACE_FOPS, 3, TRCID_CXISYSTEM_039a,
           "uiomove exit 1: rc %d data %08X %08X %08X %08X\n",
           rc, trcdata[0], trcdata[1], trcdata[2], trcdata[3]);
#else
    TRACE1(TRACE_FOPS, 3, TRCID_CXISYSTEM_039,
           "uiomove exit 1: rc %d\n",
           rc);
#endif
    return rc;
  }
  while (nBytes > 0 && uioP->uio_resid && rc == 0)
  {
    if (uioP->uio_iovcnt <= 0)
      return ENOMEM;
    iovP = uioP->uio_iov;
    cnt = iovP->iov_len;
    if (cnt <= 0)
    {
      uioP->uio_iovcnt--;
      uioP->uio_iov++;
      uioP->uio_iovdcnt++;
      continue;
    }
    if (cnt > nBytes)
      cnt = nBytes;

    if (toKernel)
    {
      /* The daemon needs to bypass access checks since copy to 
       * shared segment would inadvertantly fail.  Copies to 
       * kernel address space also perform no validity check.
       */
      if (PROCESS_GROUP(current) == DaemonPGrp || uioP->uio_segflg == UIO_SYSSPACE)
        __copy_from_user(kBufP, (char *)iovP->iov_base, cnt);
      else 
        if (copy_from_user(kBufP, (char *)iovP->iov_base, cnt))
          return EFAULT;
    }
    else
    {
      /* The daemon needs to bypass access checks since copy to 
       * shared segment would inadvertantly fail.  Copies to 
       * kernel address space also perform no validity check.
       */
      if (PROCESS_GROUP(current) == DaemonPGrp || uioP->uio_segflg == UIO_SYSSPACE)
        __copy_to_user((char *)iovP->iov_base, kBufP, cnt);
      else 
        if (copy_to_user((char *)iovP->iov_base, kBufP, cnt))
          return EFAULT;
    }
    iovP->iov_base = (char *)iovP->iov_base + cnt;
    iovP->iov_len -= cnt;
    uioP->uio_resid -= cnt;
    uioP->uio_offset += cnt;
    kBufP += cnt;
    nBytes -= cnt;
  }
#ifdef TRACE_IO_DATA
  cnt = kBufP - origKBufP;
  if (cnt >= sizeof(trcdata))
    memcpy(trcdata, origKBufP, sizeof(trcdata));
  else
  {
    memset(trcdata, 0xAA, sizeof(trcdata));
    memcpy(trcdata, origKBufP, cnt);
  }
  TRACE5(TRACE_FOPS, 3, TRCID_CXISYSTEM_041a,
         "uiomove exit 2: rc %d data %08X %08X %08X %08X\n",
         rc, trcdata[0], trcdata[1], trcdata[2], trcdata[3]);
#else
  TRACE1(TRACE_FOPS, 3, TRCID_CXISYSTEM_041,
         "uiomove exit 2: rc %d\n",
         rc);
#endif
  return rc;
}

/*
  Try to force some sanity checks at compile type
 */
/* TO DO: revise this to handle comparisons beyond equality/inequality */
/* STATIC_DBGASSERT(sizeof(spinlock_t), SPINLOCK_T_SIZE); */

/* A routine to check that the definitions in our cxiTypes.h
 * files are equivalent to the system definitions.  The module
 * should not load if it receives an error from this routine.
 */
int
cxiCheckTypes()
{
  int rc = 0;

  /* Make sure cxiBlockingMutex_t fits in the space provided.  If not,
     the implementation of the cxiBlockingMutex... routines needs to
     use the embedded space to record a pointer to kmalloc'ed space holding
     the semaphore. */
  if (sizeof(struct semaphore) > GPFS_LINUX_SEM_SIZE)
  {
    printk("cxiCheckTypes: semaphore %ld > GPFS_LINUX_SEM_SIZE %ld\n",
           sizeof(struct semaphore), GPFS_LINUX_SEM_SIZE);
    rc = 1;
  }

  /* Size of spinlock_t is smaller for UP case with gcc 3.x, so just
     insure SPINLOCK_T_SIZE is large enough for both the UP and SMP case. */
  if (sizeof(spinlock_t) > SPINLOCK_T_SIZE)
  {
    printk("cxiCheckTypes: spinlock_t %ld > SPINLOCK_T__SIZE %ld\n",
           sizeof(spinlock_t), SPINLOCK_T_SIZE);
    rc = 2;
  }

  /* Ensure that size of pid_t matches cxiThreadId (32-bits) */
  if (sizeof(pid_t) != sizeof(cxiThreadId))
  {
    printk("cxiCheckTypes: pid_t %ld != cxiThreadId %ld\n",
           sizeof(pid_t), sizeof(cxiThreadId));
    rc = 3;
  }

  if (rc > 0)
    TRACE1(TRACE_TASKING, 2, TRCID_CXISYSTEM_CHKTYPES,
           "cxiCheckTypes: system type mismatch on type number %d!\n", rc);
  return rc;
}

/* Routine to get current time of day in nanosecond format.
 */
int
cxiGetTOD(cxiTimeStruc_t *tsP)
{
  struct timeval tv;

  /* This call returns microseconds so we fudge it to nanoseconds */
  do_gettimeofday(&tv);
  tsP->tv_sec = tv.tv_sec;
  tsP->tv_nsec = tv.tv_usec * 1000;

  return 0;
}

Boolean
cxiIsNFSThread()
{
# if defined(GPFS_LITTLE_ENDIAN) && !defined(__64BIT__)
    /* Note comparison against a multibyte character constant (not a string
      constant).  Order of characters in word is reversed due to little-
      endian representation of integers. */
    if (* ((int*)&current->comm[0]) != 0x6473666e) // 'dsfn'
       return false;
    return * ((char*)&current->comm[4]) == '\0';
# else
    return (strcmp(current->comm, "nfsd") == 0);
# endif
}

Boolean
cxiIsNFS4Thread()
{
# if defined(GPFS_LITTLE_ENDIAN) && !defined(__64BIT__)
    /* Note comparison against a multibyte character constant (not a string
      constant).  Order of characters in word is reversed due to little-
      endian representation of integers. */
    if ((* ((int*)&current->comm[0]) != 0x6473666e) |  // 'dsfn'
        (* ((int*)&current->comm[2]) != 0x00346473));  // '4ds'
       return false;
    return * ((char*)&current->comm[5]) == '\0';
# else
    return (strcmp(current->comm, "nfsd4") == 0);
# endif
}

Boolean
cxiIsKupdateThread()
{
  return (strcmp(current->comm, "kupdated") == 0);
}

#ifdef SMB_LOCKS
Boolean
cxiIsSambaOrLockdThread()
{
# if defined(GPFS_LITTLE_ENDIAN) && !defined(__64BIT__)
    /* Note comparison against a multibyte character constant (not a string
      constant).  Order of characters in word is reversed due to little-
      endian representation of integers. */
    Boolean rc = (((* ((int*)&current->comm[0]) == 0x64626d73) &   // 'dbms'
                   (* ((char*)&current->comm[4]) == '\0'))    |
                  ((* ((int*)&current->comm[0]) == 0x6b636f6c) &   // 'kcol'
                   (* ((int*)&current->comm[2]) == 0x00646b63)));  // 'dkc'
       return rc;
# else
    return ((strcmp(current->comm, "smbd") == 0) |
            (strcmp(current->comm, "lockd") == 0));
# endif
}

Boolean
cxiIsSambaThread()
{
# if defined(GPFS_LITTLE_ENDIAN) && !defined(__64BIT__)
    /* Note comparison against a multibyte character constant (not a string
      constant).  Order of characters in word is reversed due to little-
      endian representation of integers. */
    Boolean rc = ((* ((int*)&current->comm[0]) == 0x64626d73) &  // 'dbms'
                  (* ((char*)&current->comm[4]) == '\0'));
       return rc;
# else
    return (strcmp(current->comm, "smbd") == 0);
# endif
}
#endif

Boolean
cxiIsGPFSThread()
{
# if defined(GPFS_LITTLE_ENDIAN) && !defined(__64BIT__)
  return (((* ((int*)&current->comm[0]) == 0x73666d6d) &  // 'sfmm'
           (* ((int*)&current->comm[2]) == 0x00647366))); // 'dsf'
# else
  return (strcmp(current->comm, "mmfsd") == 0);
# endif
}

#ifdef INSTRUMENT_LOCKS
void InitBlockingMutexStats()
{
  memset(BlockingMutexStatsTable, 0, sizeof(BlockingMutexStatsTable));
}
#endif

/* Initialize a cxiBlockingMutex_t.  Instead of the DBGASSERT, this routine
   should kmalloc a struct semaphore if bmSem is too small.  */
void cxiBlockingMutexInit(cxiBlockingMutex_t* mP, int bmNameIdx)
{
  DBGASSERT(sizeof(struct semaphore) <= GPFS_LINUX_SEM_SIZE);
#ifdef INSTRUMENT_LOCKS
  DBGASSERT(bmNameIdx < MAX_GPFS_LOCK_NAMES);
#endif  /* INSTRUMENT_LOCKS */

  TRACE2(TRACE_KLOCKL, 3, TRCID_BM_INIT,
         "cxiBlockingMutexInit: mP 0x%lX idx %d\n",
         mP, bmNameIdx);
  init_MUTEX((struct semaphore *)mP->bmSem);
  mP->bmOwnerP = NULL;
  mP->lockNameIndex = bmNameIdx;
}


/* Enter critical section, blocking this thread if necessary.  Mark this
   thread as the owner of the mutex before returning. */
void 
cxiBlockingMutexAcquire(cxiBlockingMutex_t* mP)
{
  TRACE4(TRACE_KLOCKL, 9, TRCID_BM_ACQ,
         "cxiBlockingMutexAcquire: about to acquire 0x%lX type %d "
         "current 0x%lX currentOwner 0x%lX\n",
         mP, mP->lockNameIndex, current, mP->bmOwnerP);

  DBGASSERTRC(mP->bmOwnerP != (char *)current, 
              PTR_TO_INT32(mP->bmOwnerP), PTR_TO_INT32(mP), 0);

#ifdef INSTRUMENT_LOCKS
  BlockingMutexStatsTable[mP->lockNameIndex].bmsAcquires += 1;
  if (mP->bmOwnerP != NULL)
    BlockingMutexStatsTable[mP->lockNameIndex].bmsConflicts += 1;
#endif

  down((struct semaphore *)mP->bmSem);
  mP->bmOwnerP = (char *)current;

  TRACE1(TRACE_KLOCKL, 9, TRCID_BM_ACQ_EXIT,
         "cxiBlockingMutexAcquire: returning after acquiring 0x%lX\n", mP);
}


/* Leave critical section and awaken waiting threads */
void 
cxiBlockingMutexRelease(cxiBlockingMutex_t* mP)
{
  TRACE4(TRACE_KLOCKL, 9, TRCID_BM_REL,
         "cxiBlockingMutexRelease: about to release 0x%lX type %d"
         "current 0x%lX currentOwner 0x%lX\n",
         mP, mP->lockNameIndex,current, mP->bmOwnerP);

  if (mP->bmOwnerP == (char *)current)
  {
     mP->bmOwnerP = NULL;
     up((struct semaphore *)mP->bmSem);
  }
}

/* Free resources associated with this cxiBlockingMutex_t in preparation
   for freeing the storage it occupies */
void cxiBlockingMutexTerm(cxiBlockingMutex_t* mP)
{
  TRACE2(TRACE_KLOCKL, 3, TRCID_BM_TERM,
         "cxiBlockingMutexTerm: mP 0x%lX type %d\n", mP, mP->lockNameIndex);

  /* Verify that mutex is not held */
  DBGASSERT(mP->bmOwnerP == NULL);
  DBGASSERT(atomic_read(&((struct semaphore *)mP->bmSem)->count) == 1);
}


/* Return true if a cxiBlockingMutex_t is held by the calling process */
Boolean 
cxiBlockingMutexHeldByCaller(cxiBlockingMutex_t* mP)
{
  Boolean result;
  char* ownerP;
  cxiPid_t ownerPid;

  /* Cache bmOwnerP is case it changes to NULL */
  ownerP = mP->bmOwnerP;
  if (ownerP == NULL)
    result = false;
  else
  {
    cxiThreadPtrToThreadId(ownerP, &ownerPid);
    result = (current->pid == ownerPid);
  }
  TRACE2(TRACE_KLOCKL, 9, TRCID_CXISYSTEM_017,
         "cxiBlockingMutexHeldByCaller: owner 0x%lX returns %d\n",
         ownerP, result);
  return result;
}


/* Return true if a cxiBlockingMutex_t has one or more processes waiting
   on it */
Boolean cxiBlockingMutexHasWaiters(cxiBlockingMutex_t* mP)
{
  struct semaphore * semP = (struct semaphore *)mP->bmSem;
  Boolean result;

  if ((void*)semP->wait.task_list.next != (void*)&semP->wait.task_list.next)
    result = true;
  else
    result = false;
  TRACE2(TRACE_KLOCKL, 9, TRCID_CXISYSTEM_018,
         "cxiBlockingMutexHasWaiters: mP 0x%lX hasWaiters %d\n",
         mP, result);
}


/* Wait for a cxiWaitEventSignal, cxiWaitEventBroadcast, or
   cxiWaitEventBroadcastRC.  Drop the associated cxiBlockingMutex_t
   *mutexP while waiting, and reacquire it before returning.
   If INTERRUPTIBLE is set in waitFlags, waits interruptibly;
   otherwise, waits uninterruptibly.
     Returns THREAD_INTERRUPTED if interrupted before being woken up,
   THREAD_AWAKENED, if woken up by cxiWaitEventSignal or
   cxiWaitEventBroadcast, or the result value passed to
   cxiWaitEventWakeupResult, if woken up by cxiWaitEventWakeupResult. */
int cxiWaitEventWait(cxiWaitEvent_t* weP, cxiBlockingMutex_t* mutexP,
                     int waitFlags)
{
  spinlock_t *lockP = (spinlock_t *)(weP->lword);
  unsigned long flags;
  cxiWaitElement_t waitElement;
  int count = 0;
  Boolean done;

  TRACE3(TRACE_KLOCKL, 3, TRCID_CXISYSTEM_EVENT_WAIT_ENTER,
         "cxiWaitEventWait enter: weP 0x%lX waitFlags 0x%X about to release "
         "mutex 0x%lX \n", weP, waitFlags, mutexP);

  /* Verify that caller is holding the mutex */
  DBGASSERTRC(mutexP->bmOwnerP == (char *)current, 
              PTR_TO_INT32(mutexP->bmOwnerP), PTR_TO_INT32(mutexP), 0);

  /* initialize our wait element */
  init_waitqueue_head(&waitElement.qhead);
  init_waitqueue_entry(&waitElement.qwaiter, current);
  __add_wait_queue(&waitElement.qhead, &waitElement.qwaiter);
  waitElement.wakeupRC = 0;

  /* update our task state to not running any more */
  if (waitFlags & INTERRUPTIBLE)
    current->state = TASK_INTERRUPTIBLE;
  else
    current->state = TASK_UNINTERRUPTIBLE;

  /* add our wait element to the end of the wait list */
  SPIN_LOCK_IRQ(lockP, flags);

  CXI_WAIT_LIST_ADD(&weP->waitList, &waitElement.waitList);

  SPIN_UNLOCK_IRQ(lockP, flags);

  /* Release the mutex.  Note: calling cxiBlockingMutexRelease here is
     problematic, because it makes trace calls, which may block the current
     process, which would overwrite the task state (current->state) we just
     updated.  A way around this would be to move out task state update to
     after the call to cxiBlockingMutexRelease, but then, before calling
     schedule(), we would have to re-acquire the wait-list lock and check
     wakeupRC to see whether somebody has already woken us up since we
     released the mutex.  Since there is a trace at the top of this routine,
     we don't need the one in cxiBlockingMutexRelease; hence, just do the
     release right here. */
  mutexP->bmOwnerP = NULL;
  up((struct semaphore *)mutexP->bmSem);

again:
  /* call the scheduler */
  schedule();

  /* Remove ourself from the wait list ... except:
     It appears that kswapd is special, and even though it waits
     uninterruptibly, the page_alloc code may call wakeup_kswapd().
     This causes a unexpected premature wake up in this code.  Perhaps we
     should hand off kswapd's work to another thread so that it won't ever
     block in GPFS???  It could possibly make better forward progress
     elsewhere in the kernel.  The count of 50 is an arbitrary number to allow
     us out and assert.  Hopefully the page allocation won't ever get in that
     much trouble. */
  SPIN_LOCK_IRQ(lockP, flags);

  if (waitElement.wakeupRC == 0 &&
      !(waitFlags & INTERRUPTIBLE) &&
      cxiIsKswapdThread() &&
      count++ < 50)
  {
    current->state = TASK_UNINTERRUPTIBLE;
    done = false;
  }
  else
  {
    CXI_WAIT_LIST_REMOVE(&waitElement.waitList);
    done = true;
  }

  SPIN_UNLOCK_IRQ(lockP, flags);

  if (!done)
    goto again;

  /* re-acquire the mutex */
  cxiBlockingMutexAcquire(mutexP);

  TRACE3(TRACE_KLOCKL, 9, TRCID_CXISYSTEM_EVENT_WAIT_EXIT,
         "cxiWaitEventWait exit: weP 0x%lX mutexP 0x%lX rc %d\n",
         weP, mutexP, waitElement.wakeupRC);

  /* A zero wakeup code means we were interrupted rather than woken up */
  if (waitElement.wakeupRC != 0)
    return waitElement.wakeupRC;
  else
    return THREAD_INTERRUPTED;
}

/* Wake up one thread waiting on this cxiWaitEvent_t.  Must not sleep */
void
cxiWaitEventSignal(cxiWaitEvent_t* weP)
{
  TRACE1N(TRACE_KLOCKL, 3, TRCID_CXISYSTEM_SIGNAL,
         "cxiWaitEventSignal: weP 0x%lX\n", weP);

  doWakeup(weP, wSignal, THREAD_AWAKENED); /* wake up one */
}


/* Wake up one thread waiting on this cxiWaitEvent_t.  This is the same as
   cxiWaitEventSignal(), except this routine guarantees that multiple wake
   up calls will each pick a different thread if more than one is waiting. */
void
cxiWaitEventWakeupOne(cxiWaitEvent_t* weP)
{
  TRACE1(TRACE_KLOCKL, 3, TRCID_CXISYSTEM_WAKEUP_ONE,
         "cxiWaitEventWakeupOne: weP 0x%lX\n", weP);

  doWakeup(weP, wWakeOne, THREAD_AWAKENED); /* wake up one */
}


/* Wake up all threads waiting on this cxiWaitEvent_t */
void
cxiWaitEventBroadcast(cxiWaitEvent_t* weP)
{
  TRACE1(TRACE_KLOCKL, 3, TRCID_CXISYSTEM_BROADCAST,
         "cxiWaitEventBroadcastRC: weP 0x%lX\n", weP);

  doWakeup(weP, wBroadcast, THREAD_AWAKENED); /* wake up all */
}


/* Wake up all threads waiting on this cxiWaitEvent_t and cause them to
   return rc from their cxiWaitEventWait calls. */
void
cxiWaitEventBroadcastRC(cxiWaitEvent_t* weP, int rc)
{
  TRACE2(TRACE_KLOCKL, 3, TRCID_CXISYSTEM_BROADCAST_RC,
         "cxiWaitEventBroadcastRC: weP 0x%lX rc %d\n", weP, rc);

  doWakeup(weP, wBroadcast, rc);  /* wake up all */
}

/* alloc big memory area */
void *
cxiBigMalloc(int size)
{
  void *ptr;

  ptr = vmalloc(size);

#ifdef MALLOC_DEBUG
  MallocDebugNew(ptr, size, 2);
#endif

  return ptr;
}

/* free big memory area */
void
cxiBigFree(char *ptr)
{
#ifdef MALLOC_DEBUG
  MallocDebugDelete(ptr);
#endif

  vfree(ptr);
}

#ifdef SMB_LOCKS
/* Determine if current process has this file open */
void *
cxiCheckOpen(struct cxiNode_t* cnP)
{
  int count;
  int i;
  struct file** fdList;
  struct file*  fileP;
  struct inode* inodeP;

  count = current->files->max_fds;
  fdList = current->files->fd;
  inodeP = GNP_TO_VP(cnP);

  TRACE3(TRACE_VNODE,9,TRCID_CXICHECKOPEN_ENTRY,
         "cxiCheckOpen: entry.  %d files in fd list. Checking for inode %d "
         "at 0x%x", count, inodeP->i_ino, inodeP);

  for (i=0; i<count; i++)
  {
    fileP = fdList[i];

    if (fileP)
    {
      if (fdList[i]->f_dentry->d_inode == inodeP)
      {
        TRACE1(TRACE_VNODE, 9,TRCID_CXICHECKOPEN_FOUND,
               "cxiCheckOpen: found open file. vinfoP 0x%x",
               fileP->private_data);
        return fileP->private_data;
      }
    }
  }

  return NULL;
}

int cxiBreakOplock(void *breakArgP, int oplockNew)
{
  /* On Linux, we now use its kernel oplock support.  The get_lease()
   * call is the operation to revoke conflicting leases.  Previous versions
   * of our oplock supported used IRIX oplocks and the revoke interface
   * there was based on sending the break/file information to a pipe.
   * Although we could do a similar send to oplock_sock (oplock.c)... (if
   * we were to pass the sock into vfs_ops.oplock() as the breakArgP) it
   * seems a lot cleaner to obtain the inode* from the file* and use it
   * to call the get_lease() linux oplock break routine.  For more, refer
   * to process_local_message, request_oplock_break, F_SETSIG, and
   * oplock_sock.
   */
#ifdef OPLOCK_IRIX
  if (!pfilp || !pfilp->f_op || !pfilp->f_op->write) {
    rc = EINVAL;
    goto xerror;
  }

  writeBlock.dummy_byte = 0;
  osP->os_dev = dev;
  osP->os_ino = ino;

  oldfs = get_fs();
  set_fs(get_ds());

  spin_lock_irqsave(&current->SIG_LOCK, flags);
  oldset = current->blocked;
  sigfillset(&current->blocked);
  RECALC_SIGPENDING(current);
  spin_unlock_irqrestore(&current->SIG_LOCK, flags);

  /*
   * Write one byte of zero to send a kernel break notify
   * message followed by the status structure
   */
  rc = pfilp->f_op->write(pfilp, (char *)&writeBlock, sizeof(writeBlock),
                          &pfilp->f_pos);
  if (rc < 0)
    printk("breakSMBoplock: write on pipe failed; rc %d\n", rc);
  else {
    rc = 0;
  }

  spin_lock_irqsave(&current->SIG_LOCK, flags);
  current->blocked = oldset;
  RECALC_SIGPENDING(current);
  spin_unlock_irqrestore(&current->SIG_LOCK, flags);
  set_fs(oldfs);
#endif /* OPLOCK_IRIX */

  /* O_NONBLOCK: prevents the thread from waiting for the unlock
   * (otherwise we wait up to 45 seconds for each conflict).  Since we 
   * would have to wait for Samba to call back to us to release our 
   * oplock anyway, just wait for the whole thing here (at waitforbreak).
   */

  /* Linux op to revoke conflicting leases */
  REVOKE_LEASE((struct inode *)breakArgP, 
               O_NONBLOCK|(oplockNew==smbOplockShared)? FMODE_READ: FMODE_WRITE);

  return 0;
}

DECLARE_WAIT_QUEUE_HEAD(oplock_break_queue);

/* No initialization required on Linux */
int cxiInitBreakQ() { return 0; }

/* No initialization required on Linux */
int cxiTermBreakQ() { return 0; }

/* Send the notification that the oplock break completed */
int cxiSendBreakMsg(void *ofP)
{
  /* There is only one oplock_break_queue, and no means to pass the ofP back to
   * the waiters.  This will wake all of them up and they will recheck their
   * oplock states and wait again if necessary (with a timeout).
   */
   wake_up_interruptible(&oplock_break_queue);

  TRACE1(TRACE_SMB, 3, TRCID_SEND_BREAK, "cxiSendBreakMsg: ofP 0x%lX\n", ofP);
  return 0;
}

/* Suspend the caller until either the oplock break completes, or the timeout 
 * is reached.
 */
int cxiWaitForBreak(void *fileArgP, int oplockCurrent, int timeoutSeconds)
{
  DECLARE_WAITQUEUE(wait, current);
  signed long timeout;

  TRACE3(TRACE_SMB, 5, TRCID_BREAKWAIT, 
         "cxiWaitForBreak: file 0x%lX, oplockCurrent %d timeoutSeconds %d\n",
         fileArgP, oplockCurrent, timeoutSeconds);

  add_wait_queue(&oplock_break_queue, &wait);
  timeout = timeoutSeconds * HZ;
  while (timeout > 0) {
    set_current_state(TASK_INTERRUPTIBLE);
    /* Check whether the oplock has been released or downgraded */
    if (gpfs_ops.SMBGetOplockState(fileArgP) < oplockCurrent)
      break;
    timeout = schedule_timeout(timeout);
  }
  set_current_state(TASK_RUNNING);
  remove_wait_queue(&oplock_break_queue, &wait);

  TRACE0(TRACE_SMB, 5, TRCID_BREAKWAIT_EXIT, 
         "cxiWaitForBreak exit\n");

  return 0;
}
#endif


/* Get the address of the first byte not addressible by processes */
UIntPtr cxiGetKernelBoundary()
{
#ifdef GPFS_ARCH_IA64
  /* user addresses are below region 5 (while PAGE_OFFSET is at region 7) */
  return 0xa000000000000000UL;
#elif GPFS_ARCH_I386
  return (UIntPtr) PAGE_OFFSET;
#else
  return (UIntPtr) PAGE_OFFSET;
#endif
}


/* Return true if this process holds the big kernel lock (BKL) */
Boolean cxiHoldsBKL()
{
  return current->lock_depth >= 0;
}


/* Tell the OS that this thread is involved in handling VM page-out
   requests and should not be blocked waiting for page allocation.
   Return true if successful. */
Boolean cxiSetPageoutThread()
{
  if (current->flags & PF_MEMALLOC)
    return false;
  current->flags |= PF_MEMALLOC;
  return true;
}


/* Tell the OS that this thread is no longer involved in handling VM
   page-out requests. */
void cxiClearPageoutThread()
{
  current->flags &= ~PF_MEMALLOC;
}


/* Yield the CPU to allow other processes to run */
void
cxiYield()
{
  schedule();
}

/* Linux filldir has changed signatures depending on kernel level.
 * We always pass a 64bit offset from the GPFS layer.
 */
int
cxiFillDir(void *vargP, const char *nameP, int namelen, 
           offset_t offset, ino_t ino)
{
  cxiFillDirArg_t *fillDirArgP = (cxiFillDirArg_t *)vargP;
  filldir_t fnP = (filldir_t)fillDirArgP->fnP;

  return (*fnP)(fillDirArgP->argP, nameP, namelen, 
                (loff_t)offset, ino, 0 /* DT_UNKNOWN */);
}

#ifdef DISK_LEASE_DMS

static struct timer_list DMSTimer;
static int (*DMSgetNIOsInProgressP)(void);

#define PANIC_FOR_REAL 1

static void cxiDMSExpired(unsigned long data)
{
  int nIOs = DMSgetNIOsInProgressP();
#ifdef PANIC_FOR_REAL
  if (nIOs != 0)
    panic("GPFS Deadman Switch timer has expired, and there's still "
          " %d outstanding I/O requests\n", nIOs);
#else
  TRACE2(TRACE_DLEASE, 2, TRCID_DMS_TIMER_EXPIRED,
         "DMS timer has expired.  DMSnIOsInProgress: %d, delay %lu\n",
         nIOs, data);
#endif
}

/*
  Start dead man switch, with the timeour specified by the delay
  argument (in seconds).
*/
void cxiStartDMS(int delay, int (*funcP)(void))
{
  unsigned long njiffies = delay * HZ;

  /* Only allow the daemon or other root users to make this kernel call */
  if (!cxiIsSuperUser())
    return;

  /* There can be only one timer active at any given moment */
  if (timer_pending(&DMSTimer))
    del_timer(&DMSTimer);

  init_timer(&DMSTimer);
  DMSTimer.expires = jiffies + njiffies;
  DMSTimer.function = cxiDMSExpired;
  DMSTimer.data = delay;
  /* save the pointer to nIOsInProgress to a static var */
  DMSgetNIOsInProgressP = funcP;
  add_timer(&DMSTimer);
  TRACE2(TRACE_DLEASE, 2, TRCID_DMS_STARTED,
         "DMS timer started, delay %d, time %d\n", delay, jiffies/HZ);
}

void cxiStopDMS(void)
{
  /* Only allow the daemon or other root users to make this kernel call */
  if (!cxiIsSuperUser())
    return;

  if (timer_pending(&DMSTimer))
    del_timer(&DMSTimer);
  TRACE1(TRACE_DLEASE, 2, TRCID_DMS_STOPPED,
         "DMS timer stopped, time %d\n", jiffies/HZ);
}

/* dummy init routine.  Since on Linux the timer is
   stored in a static memory, there's nothing to be done
*/
int cxiInitDMS(void)
{
  return 0;
}

void cxiShutdownDMS(void)
{
  cxiStopDMS();
}

#endif /* DISK_LEASE_DMS */

void cxiSetBit(unsigned long *flagP, int flag_bit)
{
   set_bit(flag_bit,flagP);
}
void cxiClearBit(unsigned long *flagP, int flag_bit)
{
   clear_bit(flag_bit,flagP);
}
Boolean cxiTestBit(unsigned long *flagP, int flag_bit)
{
   return test_bit(flag_bit,flagP);
}

/* In order to setup our termination callback routine (gpfs_f_cleanup)
 * we create a dummy file and add it to our file table.  Then, upon
 * process termination, the release file operation will be called in
 * order to close the file.  The only operation we define for this
 * dummy file is release (gpfs_f_cleanup). 
 */
int
cxiRegisterCleanup()
{
  int code = 0, rc = 0;
  struct inode *iP = NULL;
  struct file *fileP = NULL;
  struct dentry *dentryP = NULL;
  extern int cleanupFD;
  extern struct super_block *shutdownSuperP;

  /* We record the daemon's process group because certain
   * checks on cxiCopyIn/cxiCopyOut are bypassed for the daemon.
   */
  DaemonPGrp = PROCESS_GROUP(current);

  /* Make sure we only create one file */
  if (cleanupFD)
    return EEXIST;

  DBGASSERT(shutdownSuperP != NULL);

  /* Allocate an inode struct */
  iP = NEW_INODE(shutdownSuperP);
  if (!iP)
  {
    code = 1;
    rc = ENOMEM;
    goto xerror;
  }
  iP->i_mode = S_IFREG;

  /* Allocate an available file descriptor */
  cleanupFD = get_unused_fd();
  if (cleanupFD < 0)
  {
    code = 2;
    rc = ENFILE;
    goto xerror;
  }

  /* Allocate a file struct */
  fileP = get_empty_filp();
  if (!fileP)
  {
    code = 3;
    rc = ENFILE;
    goto xerror;
  }

  /* Allocate a dentry sruct */
  dentryP = dget(d_alloc_root(iP));
  if (!dentryP)
  {
    code = 4;
    rc = ENOMEM;
    goto xerror;
  }

  /* Initialize and chain our file sructure */
  fileP->f_dentry = dentryP;
  fileP->f_op     = &gpfs_cleanup_fops;
  fileP->f_flags  = O_RDONLY;
  atomic_set(&fileP->f_count, 1);

  /* Just chain it on the current root mount.  When
   * the file is closed its fput() will decrement
   * the mount count (hence the mntget here)
   */
  fileP->f_vfsmnt = mntget(current->fs->rootmnt);

  /* Install the descriptor so it gets "closed" upon our termination */
  fd_install(cleanupFD, fileP);

  /* Set FD_CLOEXEC so that forked processes (like mmfsup.scr) do not
   * inherrit this descriptor.  We want the cleanup routine to be run
   * when the last mmfsd process terminates.
   */
  FD_SET(cleanupFD, current->files->close_on_exec);

  /* Once the descriptor for this dummy file is added to our file table,
   * it is inherrited by all the processes of the daemon.  As each
   * terminates, the files->count is decremented and on the last process
   * termination all the descriptors will be closed by filp_close.
   *
   * The one catch here is that our file table is inherrited by the
   * kernel threads we start as well as user processes.  This would
   * cause a problem in that daemon termination does not include these
   * kernel threads which aren't killed until restart (and therefore
   * the file is never closed).  In order for our operation to be
   * driven at daemon termiation, we must remove the file table from
   * these kernel threads.  This is done in via cxiReparent() by 
   * the mmap pager kproc.
   */

xerror:
  TRACE4(TRACE_VNODE, 1, TRCID_CXIREGISTERCLEANUP_EXIT,
         "cxiRegisterCleanup: fd %d iP %X rc %d code %d\n", 
         cleanupFD, iP, rc, code);

  if (rc)
  {
    if (dentryP);
      dput(dentryP);

    if (cleanupFD)
      put_unused_fd(cleanupFD);

    if (fileP)
      put_filp(fileP);

    if (iP)
      iput(iP);

    cleanupFD = 0;
  }

  return rc;
}

#ifdef NFS4_ACL
/* Linux routines to be called when processing NFSv4 audit/alarm ACL entries */
int cxiAuditWrite(int numargs, ...) { return ENOSYS; }
#endif /* NFS4_ACL */

/* Currently no OS specific VFS initialization for Linux */
int 
cxiInitVFS(int vfsType)
{
  return 0;
}

#if defined(DMAPI) || (SANERGY)

void cxiPathRel(void *ndP)
{
  DBGASSERT( ndP != NULL); 
  path_release( (struct nameidata *) ndP);
  cxiFreeUnpinned(ndP);
}

int
cxiPathToVfsP(void **privVfsPP, char *kpathname, void **ndPP, void **cnPP)
{
   struct gpfsVfsData_t *privVfsP = NULL;
   struct nameidata *ndP;
   struct inode * iP;
   cxiNode_t *cnP;
   int rc = 0;
    Boolean rel = false;
   int code = 0;
   *ndPP = NULL;
   *privVfsPP = NULL;

   if (kpathname == NULL)
   {
     code = 1;
     rc = EINVAL;
     goto xerror;
   }

   ndP = (struct nameidata *)cxiMallocUnpinned(sizeof(struct nameidata));
   if (ndP == NULL)
   {
     code = 2;
     rc = ENOMEM;
     goto xerror;
   }
   rc = user_path_walk(kpathname, ndP);
   if (rc)
   {
     rc = -rc;
     code = 3;
     goto xerror;
   }
   
   iP = ndP->dentry->d_inode;
   DBGASSERT(iP != NULL);
   if (!GPFS_TYPE(iP))
   {
     code = 4;
     rc = EINVAL;
     goto xerror;
   }
   
   privVfsP = VP_TO_PVP(iP);

   if (privVfsP == NULL)
   {
     code = 5;
     rc = ENOENT;
  }
  cnP = VP_TO_CNP(iP);
  *privVfsPP = (void *)privVfsP;
  *ndPP = (void *)ndP;
  if ( cnPP != NULL)
    *cnPP = (void *)cnP;

  xerror:
    if (rc && ndP)
    {
      if (rel)
      cxiPathRel(ndP);
      else
       cxiFreeUnpinned(ndP);
     }
    return rc;
}

void
cxiSetCred(void *eCredPP)
{
 ext_cred_t *eCredP = (ext_cred_t *)eCredPP;
  setCred(eCredP);
}

#endif /* DMAPI or SANERGY */
